import {
  Children,
  Fragment,
  isValidElement,
  ReactElement,
  ReactNode,
  useCallback,
  useEffect,
  useState
} from "react";
import { Route } from "react-router-dom";
import { useLogout, usePermissions } from "../auth";
import { useSafeSetState } from "../util";
import {
  AdminChildren,
  AdminRouterStatus,
  RenderRoutesFunction
} from "./types";

function isRenderRoutesFunction(children: AdminChildren): children is RenderRoutesFunction {
  return typeof children === "function";
}

function hasIteratorProtocol(children: AdminChildren): children is Iterable<ReactNode | RenderRoutesFunction> {
  return (
    typeof Symbol !== "undefined" &&
    children != null &&
    typeof (children as any)[Symbol.iterator] === "function"
  );
}

function getRenderRoutesFunctions(children: AdminChildren): RenderRoutesFunction[] {
  if (isRenderRoutesFunction(children)) {
    return [children];
  }
  if (Array.isArray(children)) {
    return children.reduce((acc, child) => {
      if (isRenderRoutesFunction(child)) {
        acc.push(child);
      }
      return acc;
    }, []);
  }
  if (hasIteratorProtocol(children)) {
    return getRenderRoutesFunctions(Array.from(children));
  }
  return [];
}

export function useConfigureAdminRouterFromChildren(children: AdminChildren): [ReactElement[], AdminRouterStatus] {
  // Gather custom routes that were declared as direct children of AppRouter
  // e.g. Not returned from the child function (if any)
  // We need to know right away whether some resources were declared to correctly
  // initialize the status at the next stop
  const doLogout = useLogout();
  const { permissions, isPending } = usePermissions();
  const [routes, setRoutes] = useState(getRoutesFromNodes(children));
  const [status, setStatus] = useSafeSetState<AdminRouterStatus>(() => getStatus(children, routes));

  const mergeRoutes = useCallback((newRoutes: ReactElement[]) => {
    setRoutes(previous => previous.concat(newRoutes));
  }, [setRoutes]);

  if (!status) {
    throw new Error('Status should be defined');
  }

  useEffect(() => {
    let funcCounts = 0;

    const onResolve = () => {
      funcCounts -= 1;
      if (funcCounts <= 0) {
        setStatus('ready');
      }
    }

    const resolveChildFunction = async (childFunc: RenderRoutesFunction) => {
      try {
        const childrenFuncResult = childFunc(permissions);
        if ((childrenFuncResult as Promise<ReactNode>)?.then) {
          (childrenFuncResult as Promise<ReactNode>).then(
            resolvedChildren => {
              mergeRoutes(getRoutesFromNodes(resolvedChildren));
              onResolve();
            }
          );
        } else {
          mergeRoutes(getRoutesFromNodes(childrenFuncResult as ReactNode));
          onResolve();
        }
      } catch (error) {
        funcCounts -= 1;
        console.error(error);
        doLogout();
      }
    };

    const updateFromChildren = async () => {
      const functionChild = getRenderRoutesFunctions(children);
      const newRoutes = getRoutesFromNodes(children);
      mergeRoutes(newRoutes);
      setStatus(
        functionChild.length > 0
          ? 'loading'
          : newRoutes.length > 0
            ? 'ready'
            : 'empty'
      );

      if (functionChild.length) {
        funcCounts = functionChild.length;
        for (const fn of functionChild) {
          resolveChildFunction(fn);
        }
      } else {
        onResolve();
      }
    };

    if (!isPending) {
      updateFromChildren();
    }
  }, [
    children,
    doLogout,
    isPending,
    permissions,
    mergeRoutes,
    setRoutes,
    setStatus,
  ]);

  return [routes, status];
}

function getStatus(children: AdminChildren, routes: ReactNode[]): AdminRouterStatus {
  return getRenderRoutesFunctions(children).length > 0
    ? 'loading'
    : routes.length > 0
      ? 'ready'
      : 'empty';
}

/**
 * Inspect the children and return an array of routable elements
 */
function getRoutesFromNodes(children: AdminChildren): ReactElement[] {
  const routes: ReactElement[] = [];

  if (isRenderRoutesFunction(children)) {
    return routes;
  }

  if (
    !Array.isArray(children) &&
    !isValidElement(children) &&
    hasIteratorProtocol(children)
  ) {
    return getRoutesFromNodes(Array.from(children));
  }

  Children.forEach(children, node => {
    if (!isValidElement<any>(node)) {
      // Ignore non-elements. This allows people to more easily inline
      // conditionals in their route config.
      return;
    } else if (node.type === Fragment) {
      routes.push(...getRoutesFromNodes(node.props.children));
    } else if (node.type === Route) {
      routes.push(node);
    } else if (process.env.NODE_ENV !== "production") {
      // TODO 获取 node.type 的 displayName
      const name = typeof node.type === "string"
        ? node.type
        : ((node.type as any).displayName || node.type.name);
      throw new Error(
        `[${name}]  is not a <Route> component. ` +
        `All component children of <Routes> must be a <Route> or <React.Fragment>`
      );
    }
  });

  return routes;
}
